From afd9a428b58671ffa5671446cc7351a349804002 Mon Sep 17 00:00:00 2001
From: U-ACER-1424F82190\raiska <raiska@acer-1424f82190.(none)>
Date: Sun, 31 Aug 2008 22:33:27 +0300
Subject: [PATCH] Add OpenGL shader support aka GLSL

---
 modules/video_output/opengl.c |  240 ++++++++++++++++++++++++++++++++++++++---
 1 files changed, 225 insertions(+), 15 deletions(-)

diff --git a/modules/video_output/opengl.c b/modules/video_output/opengl.c
index 32d56e2..e39e005 100644
--- a/modules/video_output/opengl.c
+++ b/modules/video_output/opengl.c
@@ -98,13 +98,67 @@
     #define OPENGL_MORE_EFFECT           8
 #endif
 
+/* Shader includes */
+#ifdef __APPLE__
+	//#include <OpenGL/glu.h>
+	/* Apple has shader support in OpenGL/glext.h */
+	GLhandleARB v,f,p;
+#else
+	/* Windows and Linux need tricks for shader support */
+	//#include <GL/glu.h>
+#include <GL/glext.h>
+
+	GLhandleARB v,f,p;
+	#ifdef WIN32
+		PFNGLCREATEPROGRAMOBJECTARBPROC  glCreateProgramObjectARB ;
+		PFNGLDELETEOBJECTARBPROC         glDeleteObjectARB       ;
+		PFNGLUSEPROGRAMOBJECTARBPROC     glUseProgramObjectARB   ;
+		PFNGLCREATESHADEROBJECTARBPROC   glCreateShaderObjectARB ;
+		PFNGLSHADERSOURCEARBPROC         glShaderSourceARB        ;
+		PFNGLCOMPILESHADERARBPROC        glCompileShaderARB       ;
+		PFNGLGETOBJECTPARAMETERIVARBPROC glGetObjectParameterivARB;
+		PFNGLATTACHOBJECTARBPROC         glAttachObjectARB        ;
+		PFNGLGETINFOLOGARBPROC           glGetInfoLogARB          ;
+		PFNGLLINKPROGRAMARBPROC          glLinkProgramARB         ;
+		PFNGLGETUNIFORMLOCATIONARBPROC   glGetUniformLocationARB  ;
+		PFNGLUNIFORM4FARBPROC            glUniform4fARB           ;
+		PFNGLUNIFORM3FARBPROC            glUniform3fARB           ;
+		PFNGLUNIFORM1FARBPROC            glUniform1fARB           ;
+		PFNGLUNIFORM1IARBPROC            glUniform1iARB           ;
+
+		PFNGLGETATTRIBLOCATIONARBPROC  glGetAttribLocationARB;
+		PFNGLVERTEXATTRIB3FARBPROC     glVertexAttrib3fARB;
+
+		PFNGLVERTEXATTRIBPOINTERARBPROC      glVertexAttribPointerARB;
+
+		#else
+		/* Linux part */
+		//#include <GL/glext.h>
+		#include <GL/glx.h>
+		#include <GL/glxext.h>
+		#define glXGetProcAddress(x) (*glXGetProcAddressARB) ((const GLubyte*)x)
+		PFNGLCREATEPROGRAMOBJECTARBPROC     glCreateProgramObjectARB = NULL;
+		PFNGLCREATESHADEROBJECTARBPROC      glCreateShaderObjectARB = NULL;
+		PFNGLSHADERSOURCEARBPROC            glShaderSourceARB = NULL;
+		PFNGLCOMPILESHADERARBPROC           glCompileShaderARB = NULL;
+		PFNGLGETOBJECTPARAMETERIVARBPROC    glGetObjectParameterivARB = NULL;
+		PFNGLATTACHOBJECTARBPROC            glAttachObjectARB = NULL;
+		PFNGLGETINFOLOGARBPROC              glGetInfoLogARB = NULL;
+		PFNGLLINKPROGRAMARBPROC             glLinkProgramARB = NULL;
+		PFNGLUSEPROGRAMOBJECTARBPROC        glUseProgramObjectARB = NULL;
+		PFNGLGETUNIFORMLOCATIONARBPROC      glGetUniformLocationARB = NULL;
+		PFNGLUNIFORM1FARBPROC               glUniform1f = NULL;
+
+	#endif /* ifdef WIN32 */
+#endif  /* ifdef __APPLE__ */
+
 #ifdef OPENGL_MORE_EFFECT
     #include <math.h>
-    #ifdef __APPLE__
-        #include <OpenGL/glu.h>
-    #else
-        #include <GL/glu.h>
-    #endif
+	#ifdef __APPLE__
+		#include <OpenGL/glu.h>
+	#else
+		#include <GL/glu.h>
+	#endif
 /* 3D MODEL */
     #define CYLINDER 8
     #define TORUS     16
@@ -174,6 +228,8 @@ static void Transform    ( int, float, float, int, int, int, int, double *, doub
 #define EFFECT_TEXT N_("Effect")
 #define EFFECT_LONGTEXT N_( \
     "Several visual OpenGL effects are available." )
+#define SHADER_TEXT N_("Enable shader effects")
+#define SHADER_LONGTEXT N_("Allows you to use OpenGL GLSL shaders to adjust output video")
 
 #ifndef OPENGL_MORE_EFFECT
 static const char *const ppsz_effects[] = {
@@ -209,11 +265,13 @@ vlc_module_begin();
 
 #endif
     /* Allow opengl provider plugin selection */
-    add_string( "opengl-provider", "default", NULL, PROVIDER_TEXT, 
+    add_string( "opengl-provider", "default", NULL, PROVIDER_TEXT,
                     PROVIDER_LONGTEXT, true );
     set_callbacks( CreateVout, DestroyVout );
     add_string( "opengl-effect", "none", NULL, EFFECT_TEXT,
                  EFFECT_LONGTEXT, false );
+    add_bool( "opengl-shaders", 0, NULL, SHADER_TEXT, SHADER_LONGTEXT,
+                 true );
         change_string_list( ppsz_effects, ppsz_effects_text, 0 );
 vlc_module_end();
 
@@ -237,6 +295,7 @@ struct vout_sys_t
 
     float       f_speed;
     float       f_radius;
+    int         i_shaders;
 };
 
 /*****************************************************************************
@@ -254,6 +313,7 @@ static int CreateVout( vlc_object_t *p_this )
         return VLC_ENOMEM;
 
     var_Create( p_vout, "opengl-effect", VLC_VAR_STRING | VLC_VAR_DOINHERIT );
+    var_Create( p_vout, "opengl-shaders", VLC_VAR_BOOL | VLC_VAR_DOINHERIT );
 
     p_sys->i_index = 0;
 #ifdef __APPLE__
@@ -475,8 +535,8 @@ static int Init( vout_thread_t *p_vout )
             p_sys->i_effect ++;
         }
         if (pow(2,p_sys->i_effect) < INIFILE)
-            p_sys->i_effect = pow(2,p_sys->i_effect);
-        else if ( strcmp( val.psz_string, ppsz_effects[p_sys->i_effect]))
+            p_sys->i_effect = pow( 2, p_sys->i_effect );
+        else if ( strcmp( val.psz_string, ppsz_effects[p_sys->i_effect]) )
         {
             msg_Warn( p_vout, "no valid opengl effect provided, using "
                       "\"none\"" );
@@ -513,12 +573,162 @@ static int Init( vout_thread_t *p_vout )
         glTranslatef( 0.0, 0.0, -3.0 );
 
         float f_pov_x, f_pov_y, f_pov_z;
-        f_pov_x = var_CreateGetFloat( p_vout, "opengl-pov-x");
-        f_pov_y = var_CreateGetFloat( p_vout, "opengl-pov-y");
-        f_pov_z = var_CreateGetFloat( p_vout, "opengl-pov-z");
+        f_pov_x = var_CreateGetFloat( p_vout, "opengl-pov-x" );
+        f_pov_y = var_CreateGetFloat( p_vout, "opengl-pov-y" );
+        f_pov_z = var_CreateGetFloat( p_vout, "opengl-pov-z" );
         gluLookAt(0, 0, 0, f_pov_x, f_pov_y, f_pov_z, 0, 1, 0);
     }
 #endif
+
+    /* Check if user wants shaders and if they are usable on system */
+    //var_GetBool( p_vout, "opengl-shaders", &val ));
+    if( var_GetBool( p_vout, "opengl-shaders" ) )
+    {
+    	/* Check to see if the extensions exists */
+    	char *ext = (char*)glGetString( GL_EXTENSIONS );
+    	//If the string does not exist in the required GL_EXTENSIONS, set support to zero
+    	if( strstr( ext, "GL_ARB_shading_language_100" ) == NULL )
+    	{
+    		msg_Warn( p_vout, "OpenGL device doesn't support GL_ARB_shading_language_100 extension" );
+    		p_sys->i_shaders = 0;
+    	}
+    	else
+    	{
+    		msg_Warn( p_vout, "OpenGL device supports GL_ARB_shading_language_100 extension" );
+    		p_sys->i_shaders = 1;
+    	}
+
+    	if( strstr( ext, "GL_ARB_shader_objects" ) == NULL )
+    	{
+    	    msg_Warn( p_vout, "OpenGL device doesn't support GL_ARB_shader_objects extension" );
+    	    p_sys->i_shaders = 0;
+    	}
+    	else
+    	{
+    	    msg_Warn( p_vout, "OpenGL device supports GL_ARB_shader_objects extension" );
+    	}
+
+    	if( strstr( ext, "GL_ARB_vertex_shader" ) == NULL )
+    	{
+    		msg_Warn( p_vout, "OpenGL device doesn't support GL_ARB_vertex_shader extension" );
+    		p_sys->i_shaders = 0;
+    	}
+    	else
+    	{
+    		msg_Warn( p_vout, "OpenGL device supports GL_ARB_vertex_shader extension" );
+    	}
+
+    	if( strstr( ext, "GL_ARB_fragment_shader" ) == NULL )
+    	{
+    		msg_Warn( p_vout, "OpenGL device doesn't support GL_ARB_fragment_shader extension" );
+    		p_sys->i_shaders = 0;
+    	}
+    	else
+    	{
+    		msg_Warn( p_vout, "OpenGL device supports GL_ARB_fragment_shader extension" );
+    	}
+    }
+    else
+    { /* No shaders wanted, make sure they aren't enabled */
+    	p_sys->i_shaders = 0;
+    }
+
+    /* If shaders are enabled and supported */
+    if(p_sys->i_shaders)
+    {
+		#ifdef __APPLE__
+    	/* Mac OS X just works, don't do anything */
+		#else
+			#ifdef WIN32
+			/* Create all the functions using wglGetProcAddress */
+			glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)
+					wglGetProcAddress("glCreateProgramObjectARB");
+			glDeleteObjectARB = (PFNGLDELETEOBJECTARBPROC)
+					wglGetProcAddress("glDeleteObjectARB");
+			glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)
+					wglGetProcAddress("glUseProgramObjectARB");
+			glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)
+					wglGetProcAddress("glCreateShaderObjectARB");
+			glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)
+					wglGetProcAddress("glShaderSourceARB");
+			glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)
+					wglGetProcAddress("glCompileShaderARB");
+			glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)
+					wglGetProcAddress("glGetObjectParameterivARB");
+			glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)
+					wglGetProcAddress("glAttachObjectARB");
+			glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)
+					wglGetProcAddress("glGetInfoLogARB");
+			glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)
+					wglGetProcAddress("glLinkProgramARB");
+			glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)
+					wglGetProcAddress("glGetUniformLocationARB");
+			glUniform4fARB = (PFNGLUNIFORM4FARBPROC)
+					wglGetProcAddress("glUniform4fARB");
+			glUniform3fARB = (PFNGLUNIFORM3FARBPROC)
+					wglGetProcAddress("glUniform3fARB");
+			glUniform1fARB = (PFNGLUNIFORM1FARBPROC)
+					wglGetProcAddress("glUniform1fARB");
+			glUniform1iARB = (PFNGLUNIFORM1IARBPROC)
+					wglGetProcAddress("glUniform1iARB");
+			glGetAttribLocationARB = (PFNGLGETATTRIBLOCATIONARBPROC)
+					wglGetProcAddress("glGetAttribLocationARB");
+			glVertexAttrib3fARB = (PFNGLVERTEXATTRIB3FARBPROC)
+					wglGetProcAddress("glVertexAttrib3fARB");
+
+			glVertexAttribPointerARB = (PFNGLVERTEXATTRIBPOINTERARBPROC)
+					wglGetProcAddress("glVertexAttribPointerARB");
+
+			#else
+			/* Linux part, Create all the functions using glXGetProcAddress*/
+			glCreateProgramObjectARB = (PFNGLCREATEPROGRAMOBJECTARBPROC)
+					glXGetProcAddress("glCreateProgramObjectARB");
+			glCreateShaderObjectARB = (PFNGLCREATESHADEROBJECTARBPROC)
+    				glXGetProcAddress("glCreateShaderObjectARB");
+			glShaderSourceARB = (PFNGLSHADERSOURCEARBPROC)
+    				glXGetProcAddress("glShaderSourceARB");
+			glCompileShaderARB = (PFNGLCOMPILESHADERARBPROC)
+    				glXGetProcAddress("glCompileShaderARB");
+			glGetObjectParameterivARB = (PFNGLGETOBJECTPARAMETERIVARBPROC)
+    				glXGetProcAddress("glGetObjectParameterivARB");
+			glAttachObjectARB = (PFNGLATTACHOBJECTARBPROC)
+    				glXGetProcAddress("glAttachObjectARB");
+			glGetInfoLogARB = (PFNGLGETINFOLOGARBPROC)
+    				glXGetProcAddress("glGetInfoLogARB");
+			glLinkProgramARB = (PFNGLLINKPROGRAMARBPROC)
+    				glXGetProcAddress("glLinkProgramARB");
+			glUseProgramObjectARB = (PFNGLUSEPROGRAMOBJECTARBPROC)
+    				glXGetProcAddress("glUseProgramObjectARB");
+			glGetUniformLocationARB = (PFNGLGETUNIFORMLOCATIONARBPROC)
+    				glXGetProcAddress("glGetUniformLocationARB");
+			glUniform1f = (PFNGLUNIFORM1FARBPROC)
+    				glXGetProcAddress("glUniform1fARB");
+			#endif /* ifdef WIN32 */
+		#endif /* ifdef __APPLE__ */
+
+		/* GLSL shader programs, first vertex, then fragment */
+		const char *vsh = 	"void main(){" \
+							"gl_TexCoord[0] = gl_MultiTexCoord0; "\
+							"gl_Position=ftransform();}";
+
+		const char *fsh = 	"uniform sampler2D texture;" \
+							"void main(){ gl_FragColor =  "\
+							"1.0 - texture2D(texture,gl_TexCoord[0].st);}";
+
+		/* Initalize shader programs */
+		v = glCreateShaderObjectARB( GL_VERTEX_SHADER_ARB );
+		f = glCreateShaderObjectARB( GL_FRAGMENT_SHADER_ARB );
+		glShaderSourceARB( v, 1, &vsh, NULL );
+		glShaderSourceARB( f, 1, &fsh, NULL );
+		glCompileShaderARB( v );
+		glCompileShaderARB( f );
+		p = glCreateProgramObjectARB();
+		glAttachObjectARB( p, v );
+		glAttachObjectARB( p, f );
+		glLinkProgramARB( p );
+		glUseProgramObjectARB( p );
+    }
+
     if( p_sys->p_vout->pf_unlock )
     {
         p_sys->p_vout->pf_unlock( p_sys->p_vout );
@@ -662,7 +872,7 @@ static int Manage( vout_thread_t *p_vout )
             glMatrixMode( GL_MODELVIEW );
             glLoadIdentity();
             glTranslatef( 0.0, 0.0, -3.0 );
- 
+
             float f_pov_x, f_pov_y, f_pov_z;
             f_pov_x = var_CreateGetFloat( p_vout, "opengl-pov-x");
             f_pov_y = var_CreateGetFloat( p_vout, "opengl-pov-y");
@@ -681,7 +891,7 @@ static int Manage( vout_thread_t *p_vout )
     if (p_sys->p_vout->i_alignment != p_vout->i_alignment)
     {
         p_vout->i_changes = VOUT_CROP_CHANGE;        //to force change
-        p_sys->p_vout->i_alignment = p_vout->i_alignment;    
+        p_sys->p_vout->i_alignment = p_vout->i_alignment;
     }
     return i_ret;
 }
@@ -923,7 +1133,7 @@ static void DisplayVideo( vout_thread_t *p_vout, picture_t *p_pic )
        double d_x, d_y;
        int i_distortion = p_sys->i_effect;
        float f_p = p_sys->f_radius;
- 
+
        glEnable( VLCGL_TARGET );
        glBegin(GL_QUADS);
        for (i_i = 0; i_i < p_vout->fmt_out.i_visible_width; i_i += i_n_x)
@@ -935,7 +1145,7 @@ static void DisplayVideo( vout_thread_t *p_vout, picture_t *p_pic )
 
           int i_m;
           int i_index_max = 0;
- 
+
           for (i_j = 0; i_j < p_vout->fmt_out.i_visible_height; i_j += i_n_y)
           {
             if ( i_j == i_n_y * i_n / 2) i_n_y += p_vout->fmt_out.i_visible_height % i_n;
-- 
1.5.6.4

